#include "Ansi.hpp"

#if PLATFORM_WINDOWS
#include <windows.h>
#endif

class AnsiImpl
{
public:
    AnsiImpl() :
        m_init(false),
        m_fg(Black),
        m_bg(Black)
#if PLATFORM_WINDOWS
        ,
        m_hStdOut(GetStdHandle(STD_OUTPUT_HANDLE)),
        m_attributes(0)
#endif
    {
    }

    bool m_init;
    uint32_t m_fg;
    uint32_t m_bg;
#if PLATFORM_WINDOWS
    HANDLE m_hStdOut;
    WORD m_attributes;
#endif
};

Ansi::Ansi() :
    m_pImpl(new AnsiImpl)
{
}

Ansi::Ansi(Ansi&& other) :
    m_pImpl(std::move(other.m_pImpl))
{
    other.m_pImpl.reset(new AnsiImpl);
}

Ansi::~Ansi()
{
#if PLATFORM_WINDOWS
    if (m_pImpl->m_init)
    {
        SetConsoleTextAttribute(m_pImpl->m_hStdOut, m_pImpl->m_attributes);
    }
#endif
}

Ansi& Ansi::operator=(Ansi&& other)
{
    if (this != &other)
    {
        m_pImpl = std::move(other.m_pImpl);
        other.m_pImpl.reset(new AnsiImpl);
    }
    return *this;
}

Ansi& Ansi::changeColor(uint32_t fg, uint32_t bg, bool space)
{
    uint32_t fgOld = m_pImpl->m_fg;
    uint32_t bgOld = m_pImpl->m_bg;

    if (space && m_pImpl->m_init && bgOld == bg)
    {
        fg = fgOld;
    }

#if PLATFORM_WINDOWS
    if (!m_pImpl->m_init || fgOld != fg || bgOld != bg)
    {
        CONSOLE_SCREEN_BUFFER_INFO info;
        memset(&info, 0, sizeof(info));
        GetConsoleScreenBufferInfo(m_pImpl->m_hStdOut, &info);
        WORD attributes = info.wAttributes;
        if (!m_pImpl->m_init)
        {
            m_pImpl->m_attributes = attributes;
        }
        attributes &=
            ~(FOREGROUND_RED |
              FOREGROUND_GREEN |
              FOREGROUND_BLUE |
              FOREGROUND_INTENSITY |
              BACKGROUND_RED |
              BACKGROUND_GREEN |
              BACKGROUND_BLUE |
              BACKGROUND_INTENSITY);
        WORD foregrounds[] =
        {
            0,
            FOREGROUND_RED,
            FOREGROUND_GREEN,
            FOREGROUND_RED | FOREGROUND_GREEN,
            FOREGROUND_BLUE,
            FOREGROUND_RED | FOREGROUND_BLUE,
            FOREGROUND_GREEN | FOREGROUND_BLUE,
            FOREGROUND_RED | FOREGROUND_GREEN | FOREGROUND_BLUE
        };
        WORD backgrounds[] =
        {
            0,
            BACKGROUND_RED,
            BACKGROUND_GREEN,
            BACKGROUND_RED | BACKGROUND_GREEN,
            BACKGROUND_BLUE,
            BACKGROUND_RED | BACKGROUND_BLUE,
            BACKGROUND_GREEN | BACKGROUND_BLUE,
            BACKGROUND_RED | BACKGROUND_GREEN | BACKGROUND_BLUE
        };
        attributes |= foregrounds[fg & 0x7] | backgrounds[bg & 0x7];
        if (fg & 0x8) attributes |= FOREGROUND_INTENSITY;
        if (bg & 0x8) attributes |= BACKGROUND_INTENSITY;
        SetConsoleTextAttribute(m_pImpl->m_hStdOut, attributes);
    }
#else
    bool reset =
        !m_pImpl->m_init ||
        ((bgOld & 0x8) && !(bg & 0x8)) ||
        ((fgOld & 0x8) && !(fg & 0x8)) ||
        (fg == 0x7 && bg == 0x0 && (fgOld != 0x7 || bgOld != 0x0));

    bool bright =
        (fg & 0x8) &&
        (reset || !(fgOld & 0x8));
    bool blink =
        (bg & 0x8) &&
        (reset || !(bgOld & 0x8));
    bool foreground =
        (fg & 0x7) != (reset ? 0x7 : (fgOld & 0x7));
    bool background =
        (bg & 0x7) != (reset ? 0x0 : (bgOld & 0x7));

    const char* foregrounds[] = {"30","31","32","33","34","35","36","37"};
    const char* backgrounds[] = {"40","41","42","43","44","45","46","47"};

    if (reset || bright || blink || foreground || background)
    {
        printf("\x1b[%s%s%s%s%s%s%s%s%sm",
                reset ? "0" : "",
                reset && (bright||blink||foreground||background) ? ";" : "",
                bright ? "1" : "",
                bright && (blink||foreground||background) ? ";" : "",
                blink ? "5" : "",
                blink && (foreground||background) ? ";" : "",
                foreground ? foregrounds[fg & 0x7] : "",
                foreground && (background) ? ";" : "",
                background ? backgrounds[bg & 0x7] : ""
            );
    }
#endif

    m_pImpl->m_init = true;
    m_pImpl->m_fg = fg;
    m_pImpl->m_bg = bg;

    return *this;
}

Ansi& Ansi::clear()
{
#if PLATFORM_WINDOWS
    COORD coord;
    coord.X = 0;
    coord.Y = 0;
    CONSOLE_SCREEN_BUFFER_INFO info;
    memset(&info, 0, sizeof(info));
    GetConsoleScreenBufferInfo(m_pImpl->m_hStdOut, &info);
    DWORD numWritten = 0;
    FillConsoleOutputCharacter(m_pImpl->m_hStdOut, TCHAR(' '),
                               info.dwSize.X * info.dwSize.Y,
                               coord, &numWritten);
    FillConsoleOutputAttribute(m_pImpl->m_hStdOut, info.wAttributes,
                               info.dwSize.X * info.dwSize.Y,
                               coord, &numWritten);
    SetConsoleCursorPosition(m_pImpl->m_hStdOut, coord);
#else
    fputs("\x1b[2J\x1b[H", stdout);
#endif
    return *this;
}

Ansi& Ansi::home()
{
#if PLATFORM_WINDOWS
    COORD coord;
    coord.X = 0;
    coord.Y = 0;
    SetConsoleCursorPosition(m_pImpl->m_hStdOut, coord);
#else
    fputs("\x1b[H", stdout);
#endif
    return *this;
}

Ansi& Ansi::erase()
{
#if PLATFORM_WINDOWS
    CONSOLE_SCREEN_BUFFER_INFO info;
    memset(&info, 0, sizeof(info));
    GetConsoleScreenBufferInfo(m_pImpl->m_hStdOut, &info);
    if (info.dwCursorPosition.X < info.dwSize.X)
    {
        DWORD numWritten = 0;
        FillConsoleOutputCharacter(m_pImpl->m_hStdOut, TCHAR(' '),
                                   info.dwSize.X - info.dwCursorPosition.X,
                                   info.dwCursorPosition, &numWritten);
        FillConsoleOutputAttribute(m_pImpl->m_hStdOut, info.wAttributes,
                                   info.dwSize.X - info.dwCursorPosition.X,
                                   info.dwCursorPosition, &numWritten);
    }
#else
    fputs("\x1b[K", stdout);
#endif
    return *this;
}
