#include "ByRowPhraseSolver.hpp"
#include "PhraseInterface.hpp"
#include "StateTable.hpp"
#include <queue>

using std::string;
using std::vector;
using std::queue;

#define FAST 1

ByRowPhraseSolver::ByRowPhraseSolver() :
    m_init(false)
{
}

ByRowPhraseSolver::~ByRowPhraseSolver()
{
}

void ByRowPhraseSolver::processFrame(const PhraseInterface& pi)
{
    auto& phrases = pi.getPhrases();

    if (!m_init)
    {
        //fprintf(stderr, "Initializing\n");
        m_phraseChars.clear();
        for (uint32_t iPhrase = 0; iPhrase < phrases.size(); iPhrase++)
        {
            auto& phrase = phrases[iPhrase];
            for (uint32_t iChar = 0; iChar < phrase.size(); iChar++)
            {
                m_phraseChars.push_back(PhraseChar(iPhrase, iChar));
            }
        }
        m_phraseChars.push_back(PhraseChar((uint32_t)-1, 0));

        m_table.clear();
        m_table.resize(6);
        m_newTable.clear();
        m_newTable.resize(6);
        m_searchTable.clear();
        m_searchTable.resize(6);
        for (uint32_t iRot = 0; iRot < 6; iRot++)
        {
            m_table[iRot].resize(m_phraseChars.size() * pi.getWidth());
            m_newTable[iRot].resize(m_phraseChars.size() * pi.getWidth());
            m_searchTable[iRot].resize(pi.getWidth());
        }

        m_initTable.clear();
        m_initTable.resize(m_phraseChars.size());
        for (uint32_t iPhraseChar = 0; iPhraseChar < m_phraseChars.size(); iPhraseChar++)
        {
            m_initTable[iPhraseChar].m_score =
                (m_phraseChars[iPhraseChar].m_iChar == 0) ? 0 : (uint64_t)-1;
        }

        m_init = true;
    }

    //fprintf(stderr, "New frame\n");
    const Command commands[] = {
        Command::SW,
        Command::SE,
        Command::W,
        Command::E,
        Command::CW,
        Command::CCW
    };

    UnitState startState = pi.getStartState();
    auto& endStates = pi.getEndStates();

    int32_t width = pi.getWidth();
    int32_t height = pi.getHeight();
    int32_t y = startState.m_pos.m_y;
    uint32_t rots = pi.getRots();

    for (uint32_t rot = 0; rot < rots; rot++)
    {
        for (uint32_t iPhraseChar = 0; iPhraseChar < m_phraseChars.size(); iPhraseChar++)
        {
            for (int32_t x = 0; x < width; x++)
            {
                m_table[rot][iPhraseChar * width + x].m_score =
                    (uint64_t)-1;
                m_table[rot][iPhraseChar * width + x].m_chars.clear();
            }
        }
    }
    for (uint32_t iPhraseChar = 0; iPhraseChar < m_phraseChars.size(); iPhraseChar++)
    {
        m_table[startState.m_rot][iPhraseChar * width + (startState.m_pos.m_x - pi.getXMin(startState.m_rot))] = m_initTable[iPhraseChar];
        m_initTable[iPhraseChar].m_score = (uint64_t)-1;
        m_initTable[iPhraseChar].m_chars.clear();
    }

    bool didSomething = true;
    while (didSomething)
    {
    didSomething = false;
    //fprintf(stderr, "y = %" PRId32 "\n", y);

    for (uint32_t rot = 0; rot < rots; rot++)
    {
        for (uint32_t iPhraseChar = 0; iPhraseChar < m_phraseChars.size(); iPhraseChar++)
        {
            for (int32_t x = 0; x < width; x++)
            {
                m_newTable[rot][iPhraseChar * width + x].m_score =
                    (uint64_t)-1;
                m_newTable[rot][iPhraseChar * width + x].m_chars.clear();
            }
        }
    }

    for (uint32_t iPhraseChar = 0; iPhraseChar < m_phraseChars.size(); iPhraseChar++)
    {
        auto& phraseChar = m_phraseChars[iPhraseChar];
        uint32_t iPhrase = phraseChar.m_iPhrase;
        uint32_t iChar = phraseChar.m_iChar;
        uint32_t len = (iPhrase == (uint32_t)-1) ? 0 : phrases[iPhrase].size();
        for (uint32_t rot = 0; rot < rots; rot++)
        {
            int32_t yMin = pi.getYMin(rot);
            if (y < yMin || y >= yMin + height)
            {
                continue;
            }
            int32_t xMin = pi.getXMin(rot);
            for (int32_t x = xMin; x < xMin + width; x++)
            {
                /*
                fprintf(stderr, "iPhraseChar = %" PRIu32 " (%" PRId32 ", %" PRIu32 "), rot = %" PRIu32 ", x = %" PRId32 ")\n",
                        iPhraseChar,
                        (int32_t)iPhrase,
                        iChar,
                        rot,
                        x);
                */
                auto& rec = m_table[rot][iPhraseChar * width + (x - xMin)];
                /*
                fprintf(stderr, "  score = %" PRId64 ", chars = \"%s\"\n",
                        (int64_t)rec.m_score,
                        string(rec.m_chars.begin(), rec.m_chars.end()).c_str());
                */
                uint64_t score = rec.m_score;
                if (score == (uint64_t)-1)
                {
                    continue;
                }

                UnitState curState(x, y, rot);
                int32_t curXMin = pi.getXMin(curState.m_rot);

                for (uint32_t curRot = 0; curRot < rots; curRot++)
                {
                    for (int32_t curX = 0; curX < width; curX++)
                    {
                        auto& curRec = m_searchTable[curRot][curX];
                        curRec.m_score = (uint64_t)-1;
                        curRec.m_chars.clear();
                    }
                }
                m_searchTable[rot][x - xMin] = rec;

                bool stopWord = false;

                // First make moves based on the word we are in the middle of
                //fprintf(stderr, "Making word moves\n");

                uint32_t iCurChar = iChar;
                while (iCurChar < len)
                {
                    //fprintf(stderr, "  Making word move\n");
                    char ch = phrases[iPhrase][iCurChar];
                    iCurChar++;
                    Command command = Command::Invalid;
                    if (!charToCommand(ch, &command))
                    {
                        fprintf(stderr, "ByRowPhraseSolver: invalid char\n");
                        exit(1);
                    }

                    auto& curRec = m_searchTable[curState.m_rot][curState.m_pos.m_x - curXMin];
                    /*
                    fprintf(stderr, "    curRec (rot = %" PRIu32 ", x = %" PRId32 ") chars = \"%s\"\n",
                            curState.m_rot,
                            curState.m_pos.m_x,
                            string(curRec.m_chars.begin(), curRec.m_chars.end()).c_str());
                    */
#if FAST
                    /*
                    fprintf(stderr, "(1) About to lookup command %" PRIu32 ", state (%" PRId32 ", %" PRId32 "): %" PRIu32 "\n",
                            (uint32_t)command,
                            curState.m_pos.m_x,
                            curState.m_pos.m_y,
                            curState.m_rot);
                    */
                    UnitState newState =
                        pi.transformUnitFast(command, curState);
                    //fprintf(stderr, "    Done\n");
                    bool newStateValid = (newState.m_rot != (uint32_t)-1);
#else
                    UnitState newState =
                        pi.transformUnit(command, curState);
                    bool newStateValid =
                        pi.isValid(newState);
#endif
                    int32_t newXMin = newStateValid ? pi.getXMin(newState.m_rot) : 0;
                    if (!newStateValid)
                    {
                        if (std::find(endStates.begin(),
                                      endStates.end(),
                                      curState) != endStates.end())
                        {
                        if (iCurChar < len)
                        {
                            uint32_t iNewPhraseChar = iPhraseChar + (iCurChar - iChar);
                            auto& newRec = m_initTable[iNewPhraseChar];
                            uint64_t newScore = score;
                            if ((newRec.m_score == (uint64_t)-1) ||
                                (newScore > newRec.m_score))
                            {
                                newRec.m_score = newScore;
                                newRec.m_chars = curRec.m_chars;
                                newRec.m_chars.push_back(ch);
                                /*
                                fprintf(stderr, "    appending '%c', chars = \"%s\"\n",
                                        ch,
                                        string(newRec.m_chars.begin(), newRec.m_chars.end()).c_str());
                                */
                            }
                        }
                        else
                        {
                            for (uint32_t iNewPhraseChar = 0;
                                 iNewPhraseChar < m_phraseChars.size();
                                 iNewPhraseChar++)
                            {
                                if (m_phraseChars[iNewPhraseChar].m_iChar == 0)
                                {
                                    auto& newRec = m_initTable[iNewPhraseChar];
                                    uint64_t newScore = score + 2 * len;
                                    if ((newRec.m_score == (uint64_t)-1) ||
                                        (newScore > newRec.m_score))
                                    {
                                        newRec.m_score = newScore;
                                        newRec.m_chars = curRec.m_chars;
                                        newRec.m_chars.push_back(ch);
                                        /*
                                fprintf(stderr, "    appending '%c', chars = \"%s\"\n",
                                        ch,
                                        string(newRec.m_chars.begin(), newRec.m_chars.end()).c_str());
                                        */
                                    }
                                }
                            }
                        }
                        }
                        stopWord = true;
                        break;
                    }
                    else if (newState.m_pos.m_y > curState.m_pos.m_y)
                    {
                        if (iCurChar < len)
                        {
                            uint32_t iNewPhraseChar = iPhraseChar + (iCurChar - iChar);
                            auto& newRec = m_newTable[newState.m_rot][iNewPhraseChar * width + (newState.m_pos.m_x - newXMin)];
                            uint64_t newScore = score;
                            if ((newRec.m_score == (uint64_t)-1) ||
                                (newScore > newRec.m_score))
                            {
                                newRec.m_score = newScore;
                                newRec.m_chars = curRec.m_chars;
                                newRec.m_chars.push_back(ch);
                                /*
                                fprintf(stderr, "    !! appending '%c', chars = \"%s\"\n",
                                        ch,
                                        string(newRec.m_chars.begin(), newRec.m_chars.end()).c_str());
                                */
                                didSomething = true;
                            }
                        }
                        else
                        {
                            for (uint32_t iNewPhraseChar = 0;
                                 iNewPhraseChar < m_phraseChars.size();
                                 iNewPhraseChar++)
                            {
                                if (m_phraseChars[iNewPhraseChar].m_iChar == 0)
                                {
                                    auto& newRec = m_newTable[newState.m_rot][iNewPhraseChar * width + (newState.m_pos.m_x - newXMin)];
                                    uint64_t newScore = score + 2 * len;
                                    if ((newRec.m_score == (uint64_t)-1) ||
                                        (newScore > newRec.m_score))
                                    {
                                        newRec.m_score = newScore;
                                        newRec.m_chars = curRec.m_chars;
                                        newRec.m_chars.push_back(ch);
                                        /*
                                fprintf(stderr, "    appending '%c', chars = \"%s\"\n",
                                        ch,
                                        string(newRec.m_chars.begin(), newRec.m_chars.end()).c_str());
                                        */
                                        didSomething = true;
                                    }
                                }
                            }
                        }
                        stopWord = true;
                        break;
                    }
                    auto& newRec = m_searchTable[newState.m_rot][newState.m_pos.m_x - newXMin];
                    if (newRec.m_score != (uint64_t)-1)
                    {
                        stopWord = true;
                        break;
                    }
                    newRec = curRec;
                    newRec.m_chars.push_back(ch);
                    /*
                                fprintf(stderr, "    appending '%c', chars = \"%s\"\n",
                                        ch,
                                        string(newRec.m_chars.begin(), newRec.m_chars.end()).c_str());
                    */
                    if (iCurChar == len)
                    {
                        newRec.m_score += 2 * len;
                    }

                    curState = newState;
                    curXMin = newXMin;
                }

                if (stopWord)
                {
                    continue;
                }

                // Next make moves that are not part of a word
                //fprintf(stderr, "Making non-word moves\n");

                typedef std::pair<UnitState, TableRec> Entry;
                queue<Entry> unitQueue;
                {
                    auto& rec = m_searchTable[curState.m_rot][curState.m_pos.m_x - curXMin];
                    Entry entry;
                    entry.first = curState;
                    entry.second = rec;
                    if (entry.second.m_score == (uint64_t)-1)
                    {
                        fprintf(stderr, "Bad score\n");
                        exit(1);
                    }
                    unitQueue.push(std::move(entry));
                    rec.m_score = (uint64_t)-1;
                    rec.m_chars.clear();
                }

                while (!unitQueue.empty())
                {
                    Entry curEntry(std::move(unitQueue.front()));
                    auto& curState = curEntry.first;
                    int32_t curXMin = pi.getXMin(curState.m_rot);
                    auto& curRec = m_searchTable[curState.m_rot][curState.m_pos.m_x - curXMin];
                    unitQueue.pop();
                    /*
                    fprintf(stderr, "Popped (%" PRId32 ", %" PRId32 "): %" PRIu32 "\n",
                            curState.m_pos.m_x,
                            curState.m_pos.m_y,
                            curState.m_rot);
                    */
                    if (curRec.m_score != (uint64_t)-1)
                    {
                        //fprintf(stderr, "Continuing\n");
                        continue;
                    }
                    //fprintf(stderr, "Not continuing\n");
                    curRec = std::move(curEntry.second);
                    for (uint32_t iCommand = 0; iCommand < 6; iCommand++)
                    {
                        Command command = commands[iCommand];
                        char ch = commandToChar(command);
#if FAST
                        /*
                        fprintf(stderr, "(2) About to lookup command %" PRIu32 ", state (%" PRId32 ", %" PRId32 "): %" PRIu32 "\n",
                                (uint32_t)command,
                                curState.m_pos.m_x,
                                curState.m_pos.m_y,
                                curState.m_rot);
                        */
                        UnitState newState =
                            pi.transformUnitFast(command, curState);
                        //fprintf(stderr, "    Done\n");
                        bool newStateValid = (newState.m_rot != (uint32_t)-1);
#else
                        UnitState newState =
                            pi.transformUnit(command, curState);
                        bool newStateValid =
                            pi.isValid(newState);
#endif
                        int32_t newXMin = newStateValid ? pi.getXMin(newState.m_rot) : 0;
                        if (!newStateValid)
                        {
                            if (std::find(endStates.begin(),
                                          endStates.end(),
                                          curState) != endStates.end())
                            {
                                for (uint32_t iNewPhraseChar = 0;
                                     iNewPhraseChar < m_phraseChars.size();
                                     iNewPhraseChar++)
                                {
                                    if (m_phraseChars[iNewPhraseChar].m_iChar == 0)
                                    {
                                        auto& newRec = m_initTable[iNewPhraseChar];
                                        uint64_t newScore = curRec.m_score;
                                        if ((newRec.m_score == (uint64_t)-1) ||
                                            (newScore > newRec.m_score))
                                        {
                                            newRec.m_score = newScore;
                                            newRec.m_chars = curRec.m_chars;
                                            newRec.m_chars.push_back(ch);
                                        }
                                    }
                                }
                            }
                        }
                        else if (newState.m_pos.m_y > curState.m_pos.m_y)
                        {
                            for (uint32_t iNewPhraseChar = 0;
                                 iNewPhraseChar < m_phraseChars.size();
                                 iNewPhraseChar++)
                            {
                                if (m_phraseChars[iNewPhraseChar].m_iChar == 0)
                                {
                                    auto& newRec = m_newTable[newState.m_rot][iNewPhraseChar * width + (newState.m_pos.m_x - newXMin)];
                                    uint64_t newScore = curRec.m_score;
                                    if ((newRec.m_score == (uint64_t)-1) ||
                                        (newScore > newRec.m_score))
                                    {
                                        newRec.m_score = newScore;
                                        newRec.m_chars = curRec.m_chars;
                                        newRec.m_chars.push_back(ch);
                                        didSomething = true;
                                    }
                                }
                            }
                        }
                        else
                        {
                            auto& newRec = m_searchTable[newState.m_rot][newState.m_pos.m_x - newXMin];
                            if (newRec.m_score == (uint64_t)-1)
                            {
                                Entry entry;
                                entry.first = newState;
                                entry.second = curRec;
                                entry.second.m_chars.push_back(ch);
                                unitQueue.push(std::move(entry));
                            }
                        }
                    }
                }
            }
        }
    }

    {
        using std::swap;
        swap(m_table, m_newTable);
    }

    y++;
    }
}

void ByRowPhraseSolver::getChars(std::vector<char>* pChars)
{
    /*
    fprintf(stderr, "Power score: %" PRIu64 "\n",
            m_initTable[m_phraseChars.size() - 1].m_score);
    */
    *pChars = m_initTable[m_phraseChars.size() - 1].m_chars;
}
