// -*- mode:c++; c-style:k&r; c-basic-offset:4; indent-tabs-mode: nil; -*-
// vi:set et cin sw=4 cino=>se0n0f0{0}0^0\:0=sl1g0hspst0+sc3C0/0(0u0U0w0m0:

#include <iostream>
#include <cassert>
#include <sstream>
#include "Encoder.hpp"

Encoder::Encoder(int width, int num_lines, const Vocabulary &vocab,
                 const std::vector<const Frame> &frames)
    : m_width(width), m_height(num_lines), m_num_pieces(frames.size()),
      m_vocab(vocab), m_frame_spawns(), m_frame_final_locs(),
      m_frame_rot_lims(), m_frame_x_mins(), m_frame_y_mins(),
      m_memo(width * num_rotations * (1 + num_lines * frames.size()), unknown) {
    // faux spawn loc for final locking move
    for (auto c = 0; c < width; ++c) {
        for (auto r = 0; r < 6; ++r) {
            memo_at(num_lines * frames.size(), c, r) = invalid;
        }
    }
    memo_at(num_lines * frames.size(), 0, 0) = original;

    // actual frames
    for (auto f = 0; f < frames.size(); ++f) {
        auto &frame = frames[f];
        memo_at(f * num_lines + pos_y(frame.spawn), pos_x(frame.spawn), pos_r(frame.spawn)) = original;
        if (f == 0) memo_at(pos_y(frame.spawn), pos_x(frame.spawn), pos_r(frame.spawn)).score = 0; // score for init pt on frame 0 is known
        m_frame_spawns.push_back(frame.spawn);
        for (auto i : frame.invalid_locs) {
            memo_at(f * num_lines + pos_y(i), pos_x(i), pos_r(i)) = invalid;
        }
        std::vector<std::vector<Pos> > frame_finals;
        for (auto l : frame.final_locs) {
            std::vector<Pos> final_moves(vocab.primitives().size(), std::make_tuple(-1,-1,-1));
            int adj_x = (pos_y(l) % num_lines) == 0 ? 0 : 1;
            for (int mi = 0; mi < vocab.primitives().size(); ++mi) {
                const Delta &m = vocab.primitives()[mi].rpath[0];
                int x = pos_x(l) + pos_x(m) + (pos_y(m) == 0 ? 0 : adj_x);
                int y = pos_y(l) + pos_y(m);
                int r = pos_r(l) + pos_r(m); if (r < 0) r += 6; if (r >= 6) r -= 6;
                if (x >= 0 && x < width && y >= 0 && y < num_lines
                    && !memo_at(f * num_lines + y, x, r).is_invalid()) continue; // not a locking move
                // std::cerr << "DEBUG: init fs cmd " << m_vocab.texts()[mi] << "  s " << pos_y(l) << " x " << pos_x(l) << " r " << pos_r(l) << " -> inv s " << y << " x " << x << " r " << r << "   mx " << pos_x(m) << " my " << pos_y(m) << " adj " << adj_x << std::endl;
                final_moves[mi] = std::make_tuple(pos_x(l), f * num_lines + pos_y(l), pos_r(l));
            }
            // std::cerr << "---" << std::endl;
            frame_finals.push_back(final_moves);
        }
        m_frame_final_locs.push_back(frame_finals);
        m_frame_rot_lims.push_back(frame.rot_limit);
        assert(frame.x_mins.size() == frame.rot_limit);
        m_frame_x_mins.push_back(frame.x_mins);
        assert(frame.y_mins.size() == frame.rot_limit);
        m_frame_y_mins.push_back(frame.y_mins);
    }
    // fake single-line frame
    m_frame_rot_lims.push_back(1);
    m_frame_x_mins.push_back(std::vector<int>(1, 0));
    m_frame_y_mins.push_back(std::vector<int>(1, 0));
    // std::cerr << "-------" << std::endl;
}

Delta Encoder::adj_delta(Delta delta, int end_step) {
    if (((end_step % m_height) % 2 != 0) || (pos_y(delta) % 2 == 0)) {
        return delta;
    } else {
        return std::make_tuple(pos_x(delta)-1, pos_y(delta), pos_r(delta));
    }
}

// XXX change to set?
std::vector<Pos>
Encoder::validate_move(int cur_x, int cur_rot, int cur_step, const Word &w) {
    return validate_move_rec(cur_x, cur_rot, cur_step, w.rpath.cbegin(), w.rpath.cend());
}

std::vector<Pos>
Encoder::validate_move_rec(int cur_x, int cur_rot, int cur_step,
                           std::vector<const Delta>::const_iterator d_cur,
                           std::vector<const Delta>::const_iterator d_end) {
    if (d_cur == d_end) {
        // std::cerr << "DEBUG: vm ret s " << cur_step << " x " << cur_x << " r " << cur_rot << std::endl;
        return std::vector<Pos>(1, std::make_tuple(cur_x, cur_step, cur_rot));
    }
    int dx = 0, dr = 0, dy = 0;
    if (memo_at(cur_step, cur_x, cur_rot).is_orig()) {
        int prev_frame_no = cur_step / m_height - 1;
        // std::cerr << "DEBUG: vm orig s " << cur_step << " x " << cur_x << " r " << cur_rot << std::endl;
        if (prev_frame_no < 0) return std::vector<Pos>(); // word extends past spawn
        std::vector<Pos> vs;
        assert(prev_frame_no < m_frame_final_locs.size());
        auto &final_locs = m_frame_final_locs[prev_frame_no];
        for (auto &locs : final_locs) {
            // XXX replace loop/if to only use that one command
            // XXX (requires an indirection layer in the vocab data struct)
            for (int cmd = 0; cmd < locs.size(); ++cmd) {
                if ((m_vocab.words()[cmd].rpath[0] == *d_cur)
                    && (pos_x(locs[cmd]) >= 0)) { // XXX only works b/c cmd must be a prim
                    // std::cerr << "DEBUG: vmf cmd " << cmd << "  s " << pos_y(locs[cmd]) << " x " << pos_x(locs[cmd]) << " r " << pos_r(locs[cmd]) << std::endl;
                    auto &l = locs[cmd];
                    auto vrs = validate_move_rec(pos_x(l), pos_r(l), pos_y(l), d_cur + 1, d_end);

                    for (auto &vr : vrs) vs.push_back(vr);
                }
            }
        }
        // std::cerr << "DEBUG: #vs: " << vs.size() << std::endl;
        return vs;
    } else if ((cur_step % m_height == 0) && (pos_y(*d_cur) != 0)) {
        // std::cerr << "DEBUG: vm cross frame s " << cur_step << " x " << cur_x << " r " << cur_rot << std::endl;
        return std::vector<Pos>(); // frame crossing @ non-spawn location
    } else {
        // std::cerr << "DEBUG: vm std s " << cur_step << " x " << cur_x << " r " << cur_rot << std::endl;
        auto d = adj_delta(*d_cur, cur_step);
        dx += pos_x(d);
        dy += pos_y(d);
        dr = (dr + pos_r(d)); if (dr < 0) dr += 6; if (dr >= 6) dr -= 6;
        cur_x += pos_x(d);
        cur_step += pos_y(d);
        cur_rot = (cur_rot + pos_r(d)); if (cur_rot < 0) cur_rot += 6; if (cur_rot >= 6) cur_rot -= 6;
        if ((cur_x < 0) || (cur_x >= m_width) || (cur_step < 0) || (cur_step >= m_height * m_num_pieces)) {
            // std::cerr << "DEBUG: vm std BOUNDS s " << cur_step << " x " << cur_x << " r " << cur_rot << std::endl;
            return std::vector<Pos>();
        }
        auto &m = memo_at(cur_step, cur_x, cur_rot);
        if (m.is_invalid() || !m.is_known()) {
            // std::cerr << "DEBUG: vm std INV OR KNOWN s " << cur_step << " x " << cur_x << " r " << cur_rot << std::endl;
            return std::vector<Pos>();
        }
        ++d_cur;
        return validate_move_rec(cur_x, cur_rot, cur_step, d_cur, d_end);
    }
}

void Encoder::compute_memo_table() {
    for (auto step = 0; step < m_height * m_num_pieces + 1; ++step) {
        int frame_no = step / m_height;
        auto r_lim = m_frame_rot_lims[frame_no];
        // std::cerr << "STEP " << step << std::endl;
        for (auto r0 = 0; r0 < r_lim; ++r0) {
            auto r = (r0 + pos_r(m_frame_spawns[frame_no])) % r_lim;
            auto x_min = m_frame_x_mins[frame_no][r];
            auto y_min = m_frame_y_mins[frame_no][r];
            if (step % m_height < y_min) continue;
            for (auto x0 = 0; x0 < m_width; ++x0) {
                auto x = (x0 + pos_x(m_frame_spawns[frame_no])) % m_width;
                if (x < x_min) continue;
                auto &cur = memo_at(step, x, r);
                std::cerr << "CUR: s " << step << " x " << x << " r " << r << " k " << cur.is_known() << " v " << !cur.is_invalid() << std::endl;
                if (cur.is_known() || cur.is_invalid()) continue;
                for (int m = 0; m < m_vocab.size(); ++m) {
                    // std::cerr << "MV: m " << m << " (\"" << m_vocab.texts()[m] << "\")" << std::endl;
                    auto word = m_vocab.words()[m];
                    auto vs = validate_move(x, r, step, word);
                    for (auto v : vs) { // valid positions after move
                        auto prev = memo_at(pos_y(v), pos_x(v), pos_r(v));
                        auto new_score = prev.score + word.score;
                        if (new_score > cur.score) {
                            std::cerr << "UPD: s " << step << " x " << x << " r " << r << " sc " << new_score << " (\"" << m_vocab.texts()[m] << "\") <UPDATED>" << " <- s " << pos_y(v) << " x " << pos_x(v) << " r " << pos_r(v) << " sc " << prev.score << std::endl;
                            cur.score = new_score;
                            cur.move_no = m;
                            cur.prev = v;
                        } else {
                            std::cerr << "N/U: s " << step << " x " << x << " r " << r << " sc " << new_score << " (\"" << m_vocab.texts()[m] << "\")" << " <- s " << pos_y(v) << " x " << pos_x(v) << " r " << pos_r(v) << " sc " << prev.score << std::endl;
                        }
                    }
                }
            }
        }
    }
}

std::tuple<int,std::string> Encoder::trace_solution() {
    auto last_step = m_height * m_num_pieces;
    int best_x = 0, best_r = 0, best_score = memo_at(last_step, 0, 0).score;
    for (auto x = 0; x < m_width; ++x) {
        for (auto r = 0; r < 6; ++r) {
            auto memo = memo_at(last_step, x, r);
            if (!memo.is_invalid() && (memo.score > best_score)) {
                best_x = x; best_r = r; best_score = memo.score;
            }
        }
    }
    std::cerr << "BEST sc " << best_score << " at x " << best_x << " r " << best_r << " step " << last_step << std::endl;

    std::vector<std::string> record;
    for (auto x = best_x, r = best_r, step = last_step; !((step / m_height == 0) && (memo_at(step,x,r).is_orig()));) {
        auto memo = memo_at(step, x, r);
        std::cerr << "STEP " << step << "  x " << x << " r " << r << " <- s " << pos_y(memo.prev) << " x " << pos_x(memo.prev) << " r " << pos_r(memo.prev) << std::endl;
        record.push_back(m_vocab.texts()[memo.move_no]);
        x = pos_x(memo.prev); r = pos_r(memo.prev); step = pos_y(memo.prev);
    }
    std::cerr << "RECS " << record.size() << std::endl;
    std::ostringstream ss;
    for (auto r = record.rbegin(); r != record.rend(); ++r) {
        ss << *r;
    }
    return std::make_tuple(best_score, ss.str());
}

std::tuple<int,std::string> Encoder::run() {
    compute_memo_table();
    return trace_solution();
}

