#include "Game.hpp"
#include "Problem.hpp"
#include "Ansi.hpp"
#include "Transform.hpp"
#include "Move.hpp"
#include <limits>
#include <algorithm>

using std::string;
using std::vector;

Game::Game() :
    m_done(false),
    m_checkVisitedStates(true),
    m_repeatedStates(false),
    m_sourceLength(0),
    m_spawnCount(0),
    m_score(0),
    m_moveScore(0),
    m_powerScore(0),
    m_linesOld(0)
{
}

void Game::init(const Problem& problem,
                const vector<string>& phrases,
                uint32_t seed)
{
    m_phrases.clear();
    for (auto& phrase : phrases)
    {
        bool ok = true;
        string processedPhrase;
        for (auto ch : phrase)
        {
            Command command = Command::Invalid;
            bool bIllegal = false;
            if (!charToCommand(ch, &command, &bIllegal))
            {
                if (bIllegal)
                {
                    ok = false;
                    break;
                }
                else
                {
                    continue;
                }
            }
            processedPhrase.push_back(lc(ch));
        }
        if (ok && !processedPhrase.empty())
        {
            m_phrases.push_back(std::move(processedPhrase));
        }
    }

    std::sort(m_phrases.begin(), m_phrases.end(),
              [](const string& phrase1, const string& phrase2) {
                  if (phrase1.size() != phrase2.size())
                  {
                      return phrase1.size() < phrase2.size();
                  }
                  return phrase1 < phrase2;
              });

    m_chars.clear();
    m_phraseCounts.clear();
    m_phraseCounts.resize(m_phrases.size());

    m_grid.resize(problem.m_width, problem.m_height);

    for (auto cell : problem.m_filled)
    {
        m_grid(cell.m_x, cell.m_y) = 1;
    }

    m_units = problem.m_units;
    m_unitInfo.resize(m_units.size());
    for (uint32_t unitIndex = 0; unitIndex < m_units.size(); unitIndex++)
    {
        auto& unit = m_units[unitIndex];
        auto& info = m_unitInfo[unitIndex];

        int32_t minX = std::numeric_limits<int32_t>::max();
        int32_t maxX = std::numeric_limits<int32_t>::min();
        int32_t minY = std::numeric_limits<int32_t>::max();
        for (auto member : unit.m_members)
        {
            minX = std::min(minX, member.m_x);
            maxX = std::max(maxX, member.m_x);
            minY = std::min(minY, member.m_y);
        }

        int32_t dmidX = minX + maxX;
        int32_t dmidXGrid = (int32_t)m_grid.getWidth() - 1;
        int32_t offsetX = (dmidXGrid - dmidX) >> 1;
        int32_t offsetY = -minY;

        info.m_initState = UnitState(unit.m_pivot.m_x + offsetX,
                                     unit.m_pivot.m_y + offsetY,
                                     0);

        vector<Cell> cells(unit.m_members);
        std::sort(cells.begin(), cells.end());

        uint32_t rots = 6;
        for (uint32_t iRot = 1; iRot <= 3; iRot++)
        {
            vector<Cell> curCells(unit.m_members);
            for (auto& cell : curCells)
            {
                cell = transformCell(unit.m_pivot,
                                     unit.m_pivot,
                                     iRot,
                                     cell);
            }
            std::sort(curCells.begin(), curCells.end());
            if (std::equal(cells.begin(), cells.end(), curCells.begin()))
            {
                rots = iRot;
                break;
            }
        }
        info.m_rots = rots;

        info.m_xMins.resize(rots);
        info.m_yMins.resize(rots);
        for (uint32_t iRot = 0; iRot < rots; iRot++)
        {
            int32_t rotMinX = std::numeric_limits<int32_t>::max();
            int32_t rotMinY = std::numeric_limits<int32_t>::max();
            for (auto member : unit.m_members)
            {
                Cell cell = transformCell(unit.m_pivot,
                                          unit.m_pivot,
                                          iRot,
                                          member);
                rotMinX = std::min(rotMinX, cell.m_x);
                rotMinY = std::min(rotMinY, cell.m_y);
            }
            info.m_xMins[iRot] = unit.m_pivot.m_x - rotMinX - 1;
            info.m_yMins[iRot] = unit.m_pivot.m_y - rotMinY;
        }
    }

    m_visitedStates.resize(6);
    for (uint32_t iRot = 0; iRot < 6; iRot++)
    {
        m_visitedStates[iRot].resize(m_grid.getWidth() + 2);
    }

    m_sourceLength = problem.m_sourceLength;
    m_source = Source(seed);
    m_spawnCount = 0;
    m_score = 0;
    m_moveScore = 0;
    m_powerScore = 0;
    m_linesOld = 0;
    spawn();
}

UnitState Game::transformUnit(Command command,
                              uint32_t unitIndex,
                              UnitState unitState) const
{
    switch (command)
    {
    case Command::E:
        return UnitState(unitState.m_pos.m_x + 1,
                         unitState.m_pos.m_y,
                         unitState.m_rot);
    case Command::W:
        return UnitState(unitState.m_pos.m_x - 1,
                         unitState.m_pos.m_y,
                         unitState.m_rot);
    case Command::SE:
        return UnitState(unitState.m_pos.m_x + (int32_t)((uint32_t)unitState.m_pos.m_y & 1),
                         unitState.m_pos.m_y + 1,
                         unitState.m_rot);
    case Command::SW:
        return UnitState(unitState.m_pos.m_x - 1 + (int32_t)((uint32_t)unitState.m_pos.m_y & 1),
                         unitState.m_pos.m_y + 1,
                         unitState.m_rot);
    case Command::CW:
        return UnitState(unitState.m_pos.m_x,
                         unitState.m_pos.m_y,
                         (unitState.m_rot + m_unitInfo[unitIndex].m_rots - 1) % m_unitInfo[unitIndex].m_rots);
    case Command::CCW:
        return UnitState(unitState.m_pos.m_x,
                         unitState.m_pos.m_y,
                         (unitState.m_rot + 1) % m_unitInfo[unitIndex].m_rots);
    default:
        return unitState;
    }
}

UnitState Game::untransformUnit(Command command,
                                uint32_t unitIndex,
                                UnitState unitState) const
{
    switch (command)
    {
    case Command::E: // W
        return UnitState(unitState.m_pos.m_x - 1,
                         unitState.m_pos.m_y,
                         unitState.m_rot);
    case Command::W: // E
        return UnitState(unitState.m_pos.m_x + 1,
                         unitState.m_pos.m_y,
                         unitState.m_rot);
    case Command::SE: // NW
        return UnitState(unitState.m_pos.m_x - 1 + (int32_t)((uint32_t)unitState.m_pos.m_y & 1),
                         unitState.m_pos.m_y - 1,
                         unitState.m_rot);
    case Command::SW: // NE
        return UnitState(unitState.m_pos.m_x + (int32_t)((uint32_t)unitState.m_pos.m_y & 1),
                         unitState.m_pos.m_y - 1,
                         unitState.m_rot);
    case Command::CW: // CCW
        return UnitState(unitState.m_pos.m_x,
                         unitState.m_pos.m_y,
                         (unitState.m_rot + 1) % m_unitInfo[unitIndex].m_rots);
    case Command::CCW: // CW
        return UnitState(unitState.m_pos.m_x,
                         unitState.m_pos.m_y,
                         (unitState.m_rot + m_unitInfo[unitIndex].m_rots - 1) % m_unitInfo[unitIndex].m_rots);
    default:
        return unitState;
    }
}

bool Game::isInRange(uint32_t unitIndex, UnitState unitState) const
{
    auto& info = m_unitInfo[unitIndex];
    auto x = unitState.m_pos.m_x;
    auto y = unitState.m_pos.m_y;
    auto rot = unitState.m_rot;
    return
        (x >= info.m_xMins[rot]) &&
        (x < info.m_xMins[rot] + (int32_t)m_grid.getWidth() + 2) &&
        (y >= info.m_yMins[rot]) &&
        (y < info.m_yMins[rot] + (int32_t)m_grid.getHeight());
}

bool Game::isValid(uint32_t unitIndex, UnitState unitState) const
{
    for (auto member : m_units[unitIndex].m_members)
    {
        Cell cell = transformCell(m_units[unitIndex].m_pivot,
                                  unitState.m_pos,
                                  unitState.m_rot,
                                  member);
        if (cell.m_x < 0 ||
            cell.m_x >= (int32_t)m_grid.getWidth() ||
            cell.m_y < 0 ||
            cell.m_y >= (int32_t)m_grid.getHeight() ||
            m_grid(cell.m_x, cell.m_y) != 0)
        {
            return false;
        }
    }

    return true;
}

void Game::spawn()
{
    if (m_spawnCount >= m_sourceLength)
    {
        m_done = true;
        return;
    }
    m_spawnCount++;

    m_curUnitIndex = m_source.next() % (uint32_t)m_units.size();
    m_curUnitState = m_unitInfo[m_curUnitIndex].m_initState;
    if (!isValid(m_curUnitIndex, m_curUnitState))
    {
        m_done = true;
        return;
    }

    resetVisitedStates();
    updateVisitedStates();
}

void Game::step(char ch)
{
    Command command = Command::E;
    bool bIllegal = false;
    if (!charToCommand(ch, &command, &bIllegal))
    {
        if (bIllegal)
        {
            m_done = true;
            m_score = 0;
            m_moveScore = 0;
            m_powerScore = 0;
        }
        return;
    }

    m_chars.push_back(ch);

    if (m_done)
    {
        m_score = 0;
        m_moveScore = 0;
        m_powerScore = 0;
        return;
    }

    for (size_t iPhrase = 0; iPhrase < m_phrases.size(); iPhrase++)
    {
        auto& phrase = m_phrases[iPhrase];
        if (phrase.size() <= m_chars.size())
        {
            size_t start = m_chars.size() - phrase.size();
            size_t i = 0;
            for (i = 0; i < phrase.size(); i++)
            {
                if (lc(phrase[i]) != lc(m_chars[start + i]))
                {
                    break;
                }
            }
            if (i == phrase.size())
            {
                uint64_t powerScore = 2 * phrase.size();
                if (m_phraseCounts[iPhrase] == 0)
                {
                    powerScore += 300;
                }
                m_score += powerScore;
                m_powerScore += powerScore;
                m_phraseCounts[iPhrase]++;
            }
        }
    }

    UnitState unitState =
        transformUnit(command, m_curUnitIndex, m_curUnitState);
    if (isValid(m_curUnitIndex, unitState))
    {
        if (unitState.m_pos.m_y != m_curUnitState.m_pos.m_y)
        {
            resetVisitedStates();
        }
        m_curUnitState = unitState;
        updateVisitedStates();
    }
    else
    {
        finishMove();
    }
}

void Game::step(Command command)
{
    step(commandToChar(command));
}

void Game::move(UnitState unitState)
{
    m_curUnitState = unitState;
    finishMove();
}

void Game::finishMove()
{
    uint64_t size = m_units[m_curUnitIndex].m_members.size();
    uint64_t lines = 0;
    vector<int32_t> rows;
    for (auto member : m_units[m_curUnitIndex].m_members)
    {
        Cell cell = transformCell(m_units[m_curUnitIndex].m_pivot,
                                  m_curUnitState.m_pos,
                                  m_curUnitState.m_rot,
                                  member);
        m_grid(cell.m_x, cell.m_y) = 1;
        rows.push_back(cell.m_y);
    }
    std::sort(rows.begin(), rows.end());
    rows.erase(std::unique(rows.begin(), rows.end()), rows.end());
    for (auto row : rows)
    {
        bool fullRow = true;
        for (uint32_t x = 0; x < m_grid.getWidth(); x++)
        {
            if (m_grid(x, row) == 0)
            {
                fullRow = false;
                break;
            }
        }
        if (fullRow)
        {
            for (uint32_t y = row; y > 0; y--)
            {
                for (uint32_t x = 0; x < m_grid.getWidth(); x++)
                {
                    m_grid(x, y) = m_grid(x, y - 1);
                }
            }
            for (uint32_t x = 0; x < m_grid.getWidth(); x++)
            {
                m_grid(x, 0) = 0;
            }
            lines++;
        }
    }
    uint64_t points = size + 100 * (((1 + lines) * lines) / 2);
    uint64_t linesBonus =
        (m_linesOld > 1) ?
        ((m_linesOld - 1) * points) / 10 :
        0;
    uint64_t moveScore = points + linesBonus;
    m_score += moveScore;
    m_moveScore += moveScore;
    m_linesOld = lines;
    spawn();
}

void Game::resetVisitedStates()
{
    for (uint32_t iRot = 0; iRot < 6; iRot++)
    {
        auto& vec = m_visitedStates[iRot];
        uint32_t size = vec.size();
        for (uint32_t x = 0; x < size; x++)
        {
            vec[x] = 0;
        }
    }
}

void Game::updateVisitedStates()
{
    auto& info = m_unitInfo[m_curUnitIndex];
    auto x = m_curUnitState.m_pos.m_x;
    auto rot = m_curUnitState.m_rot;

    auto& visited = m_visitedStates[rot][x - info.m_xMins[rot]];
    if (visited != 0)
    {
        m_repeatedStates = true;
        if (m_checkVisitedStates)
        {
            m_done = true;
            m_score = 0;
            m_moveScore = 0;
            m_powerScore = 0;
        }
    }
    visited = 1;
}

void Game::print(Ansi& ansi) const
{
    for (size_t y = 0; y <= m_grid.getHeight(); y++)
    {
        ansi.changeColor(LightGray, DarkGray, true);
        printf((y & 1) != 0 ? "  " : " ");
        for (size_t x = 0; x < m_grid.getWidth(); x++)
        {
            if (y == m_grid.getHeight())
            {
                ansi.changeColor(LightGray, DarkGray, true);
                printf("  ");
            }
            else
            {
                bool bPivot =
                    !m_done &&
                    (m_curUnitState.m_pos.m_x == (int32_t)x) &&
                    (m_curUnitState.m_pos.m_y == (int32_t)y);
                bool bUnit = false;
                if (!m_done)
                {
                    for (auto member : m_units[m_curUnitIndex].m_members)
                    {
                        Cell cell = transformCell(m_units[m_curUnitIndex].m_pivot,
                                                  m_curUnitState.m_pos,
                                                  m_curUnitState.m_rot,
                                                  member);
                        if (cell.m_x == (int32_t)x &&
                            cell.m_y == (int32_t)y)
                        {
                            bUnit = true;
                            break;
                        }
                    }
                }
                bool bFilled = (m_grid(x, y) != 0);
                if (bFilled)
                {
                    ansi.changeColor(Black, Yellow, false);
                }
                else if (bUnit)
                {
                    ansi.changeColor(Black, LightCyan, false);
                }
                else
                {
                    ansi.changeColor(DarkGray, Black, false);
                }
                printf(bPivot ? "()" : "<>");
            }
        }
        ansi.changeColor(LightGray, DarkGray, true);
        printf((y & 1) != 0 ? " " : "  ");
        ansi.changeColor(LightGray, Black, false);
        printf("\n");
    }
    printf("Score: %" PRIu64 " (%" PRIu64 " + %" PRIu64 ")\n",
           m_score, m_moveScore, m_powerScore);
}
