#include "GreedyPhraseSolver.hpp"
#include "PhraseInterface.hpp"
#include "StateTable.hpp"
#include <queue>

using std::string;
using std::vector;
using std::set;
using std::map;
using std::queue;

GreedyPhraseSolver::GreedyPhraseSolver() :
    m_init(false)
{
}

GreedyPhraseSolver::~GreedyPhraseSolver()
{
}

namespace
{
    class TableRec
    {
    public:
        TableRec() : m_prevState(Cell(), (uint32_t)-1), m_char('\0') {}

    public:
        UnitState m_prevState;
        char m_char;
    };

    uint32_t getCharsPerDescent(const string& phrase)
    {
        uint32_t chars = phrase.size();
        uint32_t descent = 0;
        for (uint32_t i = 0; i < phrase.size(); i++)
        {
            char ch = phrase[i];
            Command command = Command::Invalid;
            if (!charToCommand(ch, &command))
            {
                return 0;
            }
            if (i == phrase.size() - 1 ||
                command == Command::SW ||
                command == Command::SE)
            {
                descent++;
            }
        }
        if (descent == 0)
        {
            descent = 1;
        }
        return (chars * 1000) / descent;
    }
}

void GreedyPhraseSolver::processFrame(const PhraseInterface& pi)
{
    if (!m_init)
    {
        m_usedMap.clear();
        m_chars.clear();
        m_init = true;
    }

    auto phrases = pi.getPhrases();
    std::sort(phrases.begin(), phrases.end(),
              [&](const string& phrase1, const string& phrase2) -> bool
              {
                  bool used1 = m_usedMap[phrase1];
                  bool used2 = m_usedMap[phrase2];
                  if (used1 != used2)
                  {
                      return used2;
                  }
                  uint32_t cpd1 = getCharsPerDescent(phrase1);
                  uint32_t cpd2 = getCharsPerDescent(phrase2);
                  if (cpd1 != cpd2)
                  {
                      return cpd1 > cpd2;
                  }
                  size_t size1 = phrase1.size();
                  size_t size2 = phrase2.size();
                  if (size1 != size2)
                  {
                      return size1 < size2;
                  }
                  return phrase1 < phrase2;
              });
    /*
    fprintf(stderr, "\n");
    for (auto& phrase : phrases)
    {
        fprintf(stderr, "%s\n", phrase.c_str());
    }
    */

    const Command commands[] = {
        Command::SW,
        Command::SE,
        Command::W,
        Command::E,
        Command::CW,
        Command::CCW
    };

    UnitState startState = pi.getStartState();
    auto& endStates = pi.getEndStates();

    StateTable<TableRec> table;
    table.init(pi.m_game.m_grid.getWidth(),
               pi.m_game.m_grid.getHeight(),
               pi.getRots(),
               pi.m_game.m_unitInfo[pi.m_game.m_curUnitIndex].m_xMins,
               pi.m_game.m_unitInfo[pi.m_game.m_curUnitIndex].m_yMins);

    for (uint32_t rot = 0; rot < pi.getRots(); rot++)
    {
        for (int32_t y = 0; y < pi.getHeight(); y++)
        {
            int32_t yReal = y + pi.getYMin(rot);
            for (int32_t x = 0; x < pi.getWidth(); x++)
            {
                int32_t xReal = x + pi.getXMin(rot);
                table(xReal, yReal, rot).m_prevState.m_rot = (uint32_t)-1;
            }
        }
    }

    typedef std::pair<UnitState, TableRec> Entry;
    queue<Entry> unitQueue;

    for (auto endState : endStates)
    {
        for (uint32_t iCommand = 0; iCommand < 6; iCommand++)
        {
            Command command = commands[iCommand];
            UnitState afterEndState =
                pi.transformUnit(command, endState);
            if (!pi.isValid(afterEndState))
            {
                Entry entry;
                entry.first = endState;
                entry.second.m_prevState = endState;
                entry.second.m_char = commandToChar(command);
                unitQueue.push(std::move(entry));
                break;
            }
        }
    }

    while (!unitQueue.empty())
    {
        Entry curEntry(std::move(unitQueue.front()));
        auto& curState = curEntry.first;
        auto& curRec = curEntry.second;
        unitQueue.pop();
        if (table(curState).m_prevState.m_rot != (uint32_t)-1)
        {
            continue;
        }
        table(curState) = curRec;
        for (uint32_t iCommand = 0; iCommand < 6; iCommand++)
        {
            Command command = commands[iCommand];
            UnitState newState =
                pi.untransformUnit(command, curState);
            if (pi.isValid(newState))
            {
                if (table(newState).m_prevState.m_rot == (uint32_t)-1)
                {
                    Entry entry;
                    entry.first = newState;
                    entry.second.m_prevState = curState;
                    entry.second.m_char = commandToChar(command);
                    unitQueue.push(std::move(entry));
                }
            }
        }
    }

    vector<char> chars;

    UnitState state = startState;

    while (std::find(endStates.begin(),
                     endStates.end(),
                     state) == endStates.end())
    {

        bool foundPhrase = false;
        for (uint32_t iPhrase = 0; iPhrase < phrases.size(); iPhrase++)
        {
            auto& phrase = phrases[iPhrase];

            UnitState phraseState = state;

            set<UnitState> visitedStates;
            visitedStates.insert(phraseState);

            vector<char> phraseChars;

            bool ok = true;
            Command command = Command::Invalid;
            for (uint32_t iChar = 0; iChar < phrase.size(); iChar++)
            {
                char ch = phrase[iChar];
                phraseChars.push_back(ch);

                bool bIllegal = false;
                if (!charToCommand(ch, &command, &bIllegal))
                {
                    ok = false;
                    break;
                }

                UnitState nextState = pi.transformUnit(command, phraseState);
                if (!pi.isValid(nextState))
                {
                    ok = false;
                    break;
                }

                if (visitedStates.find(nextState) != visitedStates.end())
                {
                    ok = false;
                    break;
                }

                if (table(nextState).m_prevState.m_rot == (uint32_t)-1)
                {
                    ok = false;
                    break;
                }

                visitedStates.insert(nextState);
                phraseState = nextState;
            }
            if (!ok)
            {
                continue;
            }

            if (command != Command::SW && command != Command::SE)
            {
                ok = false;
                for (uint32_t iCommand = 0; iCommand < 2; iCommand++)
                {
                    command = commands[iCommand];
                    
                    UnitState nextState = pi.transformUnit(command, phraseState);
                    if (!pi.isValid(nextState))
                    {
                        continue;
                    }

                    if (table(nextState).m_prevState.m_rot == (uint32_t)-1)
                    {
                        continue;
                    }

                    phraseChars.push_back(commandToChar(command));
                    phraseState = nextState;
                    ok = true;
                    break;
                }
            }
            if (!ok)
            {
                continue;
            }

            state = phraseState;
            chars.insert(chars.end(), phraseChars.begin(), phraseChars.end());
            foundPhrase = true;
            m_usedMap[phrase] = true;
            break;
        }

        if (foundPhrase)
        {
            continue;
        }

        while (std::find(endStates.begin(),
                         endStates.end(),
                         state) == endStates.end())
        {
            char ch = table(state).m_char;
            chars.push_back(table(state).m_char);
            state = table(state).m_prevState;
            Command command = Command::Invalid;
            if (charToCommand(ch, &command))
            {
                if (command == Command::SW ||
                    command == Command::SE)
                {
                    break;
                }
            }
        }
    }
    chars.push_back(table(state).m_char);

    m_chars.insert(m_chars.end(), chars.begin(), chars.end());
}

void GreedyPhraseSolver::getChars(std::vector<char>* pChars)
{
    *pChars = m_chars;
}
