#include "Parser.hpp"
#include "Problem.hpp"
#include "Solution.hpp"
#include "FileUtils.hpp"
#include <jansson.h>

using std::string;
using std::vector;

bool parseCell(json_t* jCell, Cell* pCell, string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    Cell& cell = *pCell;

    if (!json_is_object(jCell))
    {
        *pMsg = "Cell not object";
        return false;
    }

    {
        json_t* jX = json_object_get(jCell, "x");
        if (!jX)
        {
            *pMsg = "Error getting x";
            return false;
        }
        if (!json_is_integer(jX))
        {
            *pMsg = "X not integer";
            return false;
        }
        cell.m_x = (int32_t)json_integer_value(jX);
    }

    {
        json_t* jY = json_object_get(jCell, "y");
        if (!jY)
        {
            *pMsg = "Error getting y";
            return false;
        }
        if (!json_is_integer(jY))
        {
            *pMsg = "Y not integer";
            return false;
        }
        cell.m_y = (int32_t)json_integer_value(jY);
    }

    return true;
}

bool parseUnit(json_t* jUnit, Unit* pUnit, string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    Unit& unit = *pUnit;

    if (!json_is_object(jUnit))
    {
        *pMsg = "Unit not object";
        return false;
    }

    {
        json_t* jMembers = json_object_get(jUnit, "members");
        if (!jMembers)
        {
            *pMsg = "Error getting members";
            return false;
        }
        if (!json_is_array(jMembers))
        {
            *pMsg = "Members not array";
            return false;
        }
        unit.m_members.clear();
        uint32_t memberCount = (uint32_t)json_array_size(jMembers);
        for (uint32_t iMember = 0; iMember < memberCount; iMember++)
        {
            json_t* jMember = json_array_get(jMembers, iMember);
            if (!jMember)
            {
                *pMsg = "Error getting member";
                return false;
            }
            Cell member;
            if (!parseCell(jMember, &member, pMsg))
            {
                return false;
            }
            unit.m_members.push_back(member);
        }
    }

    {
        json_t* jPivot = json_object_get(jUnit, "pivot");
        if (!jPivot)
        {
            *pMsg = "Error getting pivot";
            return false;
        }
        Cell pivot;
        if (!parseCell(jPivot, &pivot, pMsg))
        {
            return false;
        }
        unit.m_pivot = pivot;
    }

    return true;
}

bool parseProblem(json_t* jProblem,
                  Problem* pProblem,
                  string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    Problem& problem = *pProblem;

    if (!json_is_object(jProblem))
    {
        *pMsg = "Problem not object";
        return false;
    }

    {
        json_t* jId = json_object_get(jProblem, "id");
        if (!jId)
        {
            *pMsg = "Error getting id";
            return false;
        }
        if (!json_is_integer(jId))
        {
            *pMsg = "Id not integer";
            return false;
        }
        problem.m_id = (int64_t)json_integer_value(jId);
    }

    {
        json_t* jUnits = json_object_get(jProblem, "units");
        if (!jUnits)
        {
            *pMsg = "Error getting units";
            return false;
        }
        if (!json_is_array(jUnits))
        {
            *pMsg = "Units not array";
            return false;
        }
        problem.m_units.clear();
        uint32_t unitCount = (uint32_t)json_array_size(jUnits);
        for (uint32_t iUnit = 0; iUnit < unitCount; iUnit++)
        {
            json_t* jUnit = json_array_get(jUnits, iUnit);
            if (!jUnit)
            {
                *pMsg = "Error getting unit";
                return false;
            }
            Unit unit;
            if (!parseUnit(jUnit, &unit, pMsg))
            {
                return false;
            }
            problem.m_units.push_back(unit);
        }
    }

    {
        json_t* jWidth = json_object_get(jProblem, "width");
        if (!jWidth)
        {
            *pMsg = "Error getting width";
            return false;
        }
        if (!json_is_integer(jWidth))
        {
            *pMsg = "Width not integer";
            return false;
        }
        problem.m_width = (uint32_t)json_integer_value(jWidth);
    }

    {
        json_t* jHeight = json_object_get(jProblem, "height");
        if (!jHeight)
        {
            *pMsg = "Error getting height";
            return false;
        }
        if (!json_is_integer(jHeight))
        {
            *pMsg = "Height not integer";
            return false;
        }
        problem.m_height = (uint32_t)json_integer_value(jHeight);
    }

    {
        json_t* jFilled = json_object_get(jProblem, "filled");
        if (!jFilled)
        {
            *pMsg = "Error getting filled";
            return false;
        }
        if (!json_is_array(jFilled))
        {
            *pMsg = "Filled not array";
            return false;
        }
        problem.m_filled.clear();
        uint32_t filledCount = (uint32_t)json_array_size(jFilled);
        for (uint32_t iFilled = 0; iFilled < filledCount; iFilled++)
        {
            json_t* jFilledCell = json_array_get(jFilled, iFilled);
            if (!jFilledCell)
            {
                *pMsg = "Error getting filled cell";
                return false;
            }
            Cell filledCell;
            if (!parseCell(jFilledCell, &filledCell, pMsg))
            {
                return false;
            }
            problem.m_filled.push_back(filledCell);
        }
    }

    {
        json_t* jSourceLength = json_object_get(jProblem, "sourceLength");
        if (!jSourceLength)
        {
            *pMsg = "Error getting source length";
            return false;
        }
        if (!json_is_integer(jSourceLength))
        {
            *pMsg = "Source length not integer";
            return false;
        }
        problem.m_sourceLength = (uint64_t)json_integer_value(jSourceLength);
    }

    {
        json_t* jSourceSeeds = json_object_get(jProblem, "sourceSeeds");
        if (!jSourceSeeds)
        {
            *pMsg = "Error getting source seeds";
            return false;
        }
        if (!json_is_array(jSourceSeeds))
        {
            *pMsg = "Source seeds not array";
            return false;
        }
        problem.m_sourceSeeds.clear();
        uint32_t sourceSeedsCount = (uint32_t)json_array_size(jSourceSeeds);
        for (uint32_t iSourceSeed = 0; iSourceSeed < sourceSeedsCount; iSourceSeed++)
        {
            json_t* jSourceSeed = json_array_get(jSourceSeeds, iSourceSeed);
            if (!jSourceSeed)
            {
                *pMsg = "Error getting source seed";
                return false;
            }
            if (!json_is_integer(jSourceSeed))
            {
                *pMsg = "Source seed not integer";
                return false;
            }
            problem.m_sourceSeeds.push_back((uint32_t)json_integer_value(jSourceSeed));
        }
    }

    return true;
}

bool parseProblem(const string& strProblem,
                  Problem* pProblem,
                  string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    json_error_t jerr;
    json_t* jProblem = json_loads(strProblem.c_str(), 0, &jerr);
    if (!jProblem)
    {
        *pMsg = "Load problem failed";
        return false;
    }

    if (!parseProblem(jProblem, pProblem, pMsg))
    {
        json_decref(jProblem);
        return false;
    }
        
    json_decref(jProblem);
    return true;
}

bool loadProblem(const string& strProblemFile,
                 Problem* pProblem,
                 string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    FILE* f = fopen(strProblemFile.c_str(), "rb");
    if (!f)
    {
        *pMsg = "Error opening file ";
        *pMsg += strProblemFile.c_str();
        return false;
    }

    vector<uint8_t> data;
    readFile(f, &data);
    fclose(f);

    string strProblem;
    if (!data.empty())
    {
        strProblem = string((const char*)&data[0], data.size());
    }

    if (!parseProblem(strProblem, pProblem, pMsg))
    {
        return false;
    }

    return true;
}

bool parseSolution(json_t* jSolution,
                   Solution* pSolution,
                   string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    Solution& solution = *pSolution;

    if (!json_is_object(jSolution))
    {
        *pMsg = "Solution not object";
        return false;
    }

    {
        json_t* jProblemId = json_object_get(jSolution, "problemId");
        if (!jProblemId)
        {
            *pMsg = "Error getting problem id";
            return false;
        }
        if (!json_is_integer(jProblemId))
        {
            *pMsg = "Problem id not integer";
            return false;
        }
        solution.m_problemId = (int64_t)json_integer_value(jProblemId);
    }

    {
        json_t* jSeed = json_object_get(jSolution, "seed");
        if (!jSeed)
        {
            *pMsg = "Error getting seed";
            return false;
        }
        if (!json_is_integer(jSeed))
        {
            *pMsg = "Seed not integer";
            return false;
        }
        solution.m_seed = (uint32_t)json_integer_value(jSeed);
    }

    {
        json_t* jTag = json_object_get(jSolution, "tag");
        if (jTag)
        {
            if (!json_is_string(jTag))
            {
                *pMsg = "Tag not string";
                return false;
            }
            const char* rawTag = json_string_value(jTag);
            if (!rawTag)
            {
                *pMsg = "Error getting tag string";
                return false;
            }
            solution.m_tag = rawTag;
        }
    }

    {
        json_t* jChars = json_object_get(jSolution, "solution");
        if (!jChars)
        {
            *pMsg = "Error getting chars";
            return false;
        }
        if (!json_is_string(jChars))
        {
            *pMsg = "Chars not string";
            return false;
        }
        const char* rawChars = json_string_value(jChars);
        if (!rawChars)
        {
            *pMsg = "Error getting chars string";
            return false;
        }
        solution.m_chars.insert(solution.m_chars.end(),
                                rawChars, rawChars + strlen(rawChars));
    }

    return true;
}

bool parseSolutions(json_t* jSolutions,
                    vector<Solution>* pSolutions,
                    string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    vector<Solution>& solutions = *pSolutions;

    if (!json_is_array(jSolutions))
    {
        *pMsg = "Solutions not array";
        return false;
    }
    solutions.clear();
    uint32_t solutionCount = (uint32_t)json_array_size(jSolutions);
    for (uint32_t iSolution = 0; iSolution < solutionCount; iSolution++)
    {
        json_t* jSolution = json_array_get(jSolutions, iSolution);
        if (!jSolution)
        {
            *pMsg = "Error getting solution";
            return false;
        }
        Solution solution;
        if (!parseSolution(jSolution, &solution, pMsg))
        {
            return false;
        }
        solutions.push_back(std::move(solution));
    }

    return true;
}

bool parseSolutions(const string& strSolutions,
                    vector<Solution>* pSolutions,
                    string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    json_error_t jerr;
    json_t* jSolutions = json_loads(strSolutions.c_str(), 0, &jerr);
    if (!jSolutions)
    {
        *pMsg = "Load solutions failed";
        return false;
    }

    if (!parseSolutions(jSolutions, pSolutions, pMsg))
    {
        json_decref(jSolutions);
        return false;
    }

    json_decref(jSolutions);
    return true;
}

bool loadSolutions(const string& strSolutionsFile,
                   vector<Solution>* pSolutions,
                   string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    FILE* f = fopen(strSolutionsFile.c_str(), "rb");
    if (!f)
    {
        *pMsg = "Error opening file ";
        *pMsg += strSolutionsFile.c_str();
        return false;
    }

    vector<uint8_t> data;
    readFile(f, &data);
    fclose(f);

    string strSolutions;
    if (!data.empty())
    {
        strSolutions = string((const char*)&data[0], data.size());
    }

    if (!parseSolutions(strSolutions, pSolutions, pMsg))
    {
        return false;
    }

    return true;
}

bool formatSolutions(const vector<Solution>& solutions,
                     string* pStrSolutions,
                     string* pMsg)
{
    string msg;
    if (!pMsg)
    {
        pMsg = &msg;
    }

    json_t* jSolutions = json_array();
    if (!jSolutions)
    {
        *pMsg = "Create solutions failed";
        return false;
    }

    for (auto& solution : solutions)
    {
        json_t* jSolution = json_object();
        if (!jSolution)
        {
            json_decref(jSolutions);
            *pMsg = "Create solution failed";
            return false;
        }

        if (json_object_set_new(jSolution, "problemId",
                                json_integer(solution.m_problemId)) != 0)
        {
            json_decref(jSolution);
            json_decref(jSolutions);
            *pMsg = "Set problem id failed";
            return false;
        }

        if (json_object_set_new(jSolution, "seed",
                                json_integer(solution.m_seed)) != 0)
        {
            json_decref(jSolution);
            json_decref(jSolutions);
            *pMsg = "Set seed failed";
            return false;
        }

        if (json_object_set_new(jSolution, "tag",
                                json_string(solution.m_tag.c_str())) != 0)
        {
            json_decref(jSolution);
            json_decref(jSolutions);
            *pMsg = "Set tag failed";
            return false;
        }

        string strChars;
        if (!solution.m_chars.empty())
        {
            strChars = string(&solution.m_chars[0], solution.m_chars.size());
        }

        if (json_object_set_new(jSolution, "solution",
                                json_string(strChars.c_str())) != 0)
        {
            json_decref(jSolution);
            json_decref(jSolutions);
            *pMsg = "Set chars failed";
            return false;
        }

        if (json_array_append_new(jSolutions, jSolution) != 0)
        {
            json_decref(jSolution);
            json_decref(jSolutions);
            *pMsg = "Append solution failed";
            return false;
        }
    }

    char* rawSolutions = json_dumps(jSolutions, 0);
    if (!rawSolutions)
    {
        json_decref(jSolutions);
        *pMsg = "Dump solutions failed";
        return false;
    }
    json_decref(jSolutions);

    *pStrSolutions = string(rawSolutions);
    free(rawSolutions);
    return true;
}
