#ifndef PHRASEINTERFACE_HPP
#define PHRASEINTERFACE_HPP

#include "Common.hpp"
#include "UnitState.hpp"
#include "Game.hpp"
#include "StateTable.hpp"

class PhraseInterface
{
public:
    PhraseInterface(const Game& game) : m_game(game) {}

    // Possible phrases
    const std::vector<std::string>& getPhrases() const
        { return m_game.m_phrases; }

    // Initial state
    UnitState getStartState() const
        { return m_game.m_curUnitState; }

    // Allowed final states
    const std::vector<UnitState>& getEndStates() const
        { return m_endStates; }

    // Number of legal rotations (0..rots-1)
    uint32_t getRots() const
        { return m_game.m_unitInfo[m_game.m_curUnitIndex].m_rots; }

    // Number of legal values of x
    int32_t getWidth() const
        { return (int32_t)m_game.m_grid.getWidth() + 2; }

    // Number of legal values of y
    int32_t getHeight() const
        { return (int32_t)m_game.m_grid.getHeight(); }

    // Minimum legal value of x for the given rotation
    int32_t getXMin(uint32_t rot) const
        { return m_game.m_unitInfo[m_game.m_curUnitIndex].m_xMins[rot]; }

    // Minimum legal value of y for the given rotation
    int32_t getYMin(uint32_t rot) const
        { return m_game.m_unitInfo[m_game.m_curUnitIndex].m_yMins[rot]; }

    // Is the given state in range (x and y within min/max)?
    bool isInRange(UnitState unitState) const
        { return m_game.isInRange(m_game.m_curUnitIndex, unitState); }

    // Is the given state valid (piece can legally be there)?
    bool isValid(UnitState unitState) const
        { return m_game.isValid(m_game.m_curUnitIndex, unitState); }

    // Apply a command to a state to obtain a new state
    UnitState transformUnit(Command command, UnitState unitState) const
        { return m_game.transformUnit(command,
                                      m_game.m_curUnitIndex,
                                      unitState); }

    // Undo the affect of a command on a state
    UnitState untransformUnit(Command command, UnitState unitState) const
        { return m_game.untransformUnit(command,
                                      m_game.m_curUnitIndex,
                                      unitState); }

    // Apply a command to a state to obtain a new state and check
    // whether the new state is valid at the same time, using a
    // cache for fast lookup.  Returns a UnitState with
    // m_rot == (uint32_t)-1 if the new state is invalid.
    // To enable use of this function, override
    // PhraseSolver::wantTransformCache() and return true.
    UnitState transformUnitFast(Command command, UnitState unitState) const
        { return m_transformCache[((uint32_t)command) - 1](unitState); }

public:
    const Game& m_game;
    std::vector<UnitState> m_endStates;
    std::vector<StateTable<UnitState>> m_transformCache;
};

#endif
