#include "SimplePhraseSolver.hpp"
#include "PhraseInterface.hpp"
#include "StateTable.hpp"
#include <queue>

using std::string;
using std::vector;
using std::queue;

SimplePhraseSolver::SimplePhraseSolver() :
    m_init(false)
{
}

SimplePhraseSolver::~SimplePhraseSolver()
{
}

namespace
{
    class TableRec
    {
    public:
        TableRec() : m_prevState(Cell(), (uint32_t)-1), m_char('\0') {}

    public:
        UnitState m_prevState;
        char m_char;
    };
}

void SimplePhraseSolver::processFrame(const PhraseInterface& pi)
{
    if (!m_init)
    {
        m_chars.clear();
        m_init = true;
    }

    const Command commands[] = {
        Command::SW,
        Command::SE,
        Command::W,
        Command::E,
        Command::CW,
        Command::CCW
    };

    UnitState startState = pi.getStartState();
    auto& endStates = pi.getEndStates();

    StateTable<TableRec> table;
    table.init(pi.m_game.m_grid.getWidth(),
               pi.m_game.m_grid.getHeight(),
               pi.getRots(),
               pi.m_game.m_unitInfo[pi.m_game.m_curUnitIndex].m_xMins,
               pi.m_game.m_unitInfo[pi.m_game.m_curUnitIndex].m_yMins);

    auto phrases = pi.getPhrases();
    std::reverse(phrases.begin(), phrases.end());

    for (uint32_t iPhrase = 0; iPhrase <= phrases.size(); iPhrase++)
    {
        string phrase =
            (iPhrase < phrases.size()) ? phrases[iPhrase] : string();

        for (uint32_t rot = 0; rot < pi.getRots(); rot++)
        {
            for (int32_t y = 0; y < pi.getHeight(); y++)
            {
                int32_t yReal = y + pi.getYMin(rot);
                for (int32_t x = 0; x < pi.getWidth(); x++)
                {
                    int32_t xReal = x + pi.getXMin(rot);
                    table(xReal, yReal, rot).m_prevState.m_rot = (uint32_t)-1;
                }
            }
        }

        UnitState state = startState;
        table(startState).m_prevState = startState;
        table(startState).m_char = '\0';

        bool ok = true;
        for (uint32_t iChar = 0; iChar < phrase.size(); iChar++)
        {
            char ch = phrase[iChar];

            Command command = Command::Invalid;
            bool bIllegal = false;
            if (!charToCommand(ch, &command, &bIllegal))
            {
                ok = false;
                break;
            }

            UnitState nextState = pi.transformUnit(command, state);
            if (!pi.isValid(nextState))
            {
                ok = false;
                break;
            }

            if (table(nextState).m_prevState.m_rot != (uint32_t)-1)
            {
                ok = false;
                break;
            }

            table(nextState).m_prevState = state;
            table(nextState).m_char = ch;
            state = nextState;
        }
        if (!ok)
        {
            continue;
        }

        typedef std::pair<UnitState, TableRec> Entry;
        queue<Entry> unitQueue;
        {
            auto& rec = table(state);
            Entry entry;
            entry.first = state;
            entry.second = rec;
            unitQueue.push(std::move(entry));
            rec.m_prevState.m_rot = (uint32_t)-1;
        }

        char lastChar = '\0';
        UnitState lastState;
        bool found = false;
        while (!found && !unitQueue.empty())
        {
            Entry curEntry(std::move(unitQueue.front()));
            auto& curState = curEntry.first;
            auto& curRec = curEntry.second;
            unitQueue.pop();
            if (table(curState).m_prevState.m_rot != (uint32_t)-1)
            {
                continue;
            }
            table(curState) = curRec;
            for (uint32_t iCommand = 0; iCommand < 6; iCommand++)
            {
                Command command = commands[iCommand];
                UnitState newState =
                    pi.transformUnit(command, curState);
                if (pi.isValid(newState))
                {
                    if (table(newState).m_prevState.m_rot == (uint32_t)-1)
                    {
                        Entry entry;
                        entry.first = newState;
                        entry.second.m_prevState = curState;
                        entry.second.m_char = commandToChar(command);
                        unitQueue.push(std::move(entry));
                    }
                }
                else if (std::find(endStates.begin(),
                                   endStates.end(),
                                   curState) != endStates.end())
                {
                    lastChar = commandToChar(command);
                    lastState = curState;
                    found = true;
                    break;
                }
            }
        }

        if (found)
        {
            vector<char> chars;
            chars.push_back(lastChar);

            const TableRec* pRec = &table(lastState);
            while (pRec->m_char != '\0')
            {
                chars.push_back(pRec->m_char);
                pRec = &table(pRec->m_prevState);
            }

            std::reverse(chars.begin(), chars.end());
            m_chars.insert(m_chars.end(), chars.begin(), chars.end());
            return;
        }
    }
}

void SimplePhraseSolver::getChars(std::vector<char>* pChars)
{
    *pChars = m_chars;
}
