#include "Solver.hpp"
#include "Problem.hpp"
#include "Solution.hpp"
#include "Game.hpp"
#include "Move.hpp"
#include "Moves.hpp"
#include "Utils.hpp"
#include "Transform.hpp"
#include "PhraseInterface.hpp"
#include "PhraseSolver.hpp"
#include "CalcScore.hpp"
#include "SimplePhraseSolver.hpp"
#include "GreedyPhraseSolver.hpp"
#include "ByRowPhraseSolver.hpp"

#include <algorithm>

using std::string;
using std::vector;

inline Cell getNeighbor(uint32_t iNeighbor, Cell cell)
{
    switch (iNeighbor)
    {
    case 0: // E
        return Cell(cell.m_x + 1, cell.m_y);
    case 1: // W
        return Cell(cell.m_x - 1, cell.m_y);
    case 2: // SE
        return Cell(cell.m_x + (int32_t)((uint32_t)cell.m_y & 1), cell.m_y + 1);
    case 3: // SW
        return Cell(cell.m_x - 1 + (int32_t)((uint32_t)cell.m_y & 1), cell.m_y + 1);
    case 4: // NE
        return Cell(cell.m_x + (int32_t)((uint32_t)cell.m_y & 1), cell.m_y - 1);
    case 5: // NW
        return Cell(cell.m_x - 1 + (int32_t)((uint32_t)cell.m_y & 1), cell.m_y - 1);
    default:
        return cell;
    }
}

bool solve(const Problem& problem,
           const vector<string>& phrases,
           uint32_t iPhraseSolver,
           uint32_t seed,
           uint32_t scoreFunc,
           uint32_t scoreSeed,
           uint64_t stopMS,
           Solution* pSolution,
           uint64_t* pScore)
{
    std::unique_ptr<PhraseSolver> pPhraseSolver;
    switch (iPhraseSolver)
    {
    case 1: pPhraseSolver.reset(new SimplePhraseSolver); break;
    case 2: pPhraseSolver.reset(new GreedyPhraseSolver); break;
    case 3: pPhraseSolver.reset(new ByRowPhraseSolver); break;
    default: ;
    }

    Source scoreSource(scoreSeed);
    scoreSource.next();

    Game game;
    game.init(problem, phrases, seed);

    while (!game.m_done)
    {
        if (stopMS != 0 && getTicksMS() >= stopMS)
        {
            return false;
        }

        MoveFinder moveFinder;

        vector<UnitState> moves;
        moveFinder.findMoves(game, &moves);

        size_t bestMove = 0;
        vector<int32_t> bestMoveVec;
        uint32_t sameCount = 0;
        for (uint32_t i = 0; i < moves.size(); i++)
        {
            UnitState move = moves[i];

            int32_t curMoveYTotal = 0;
            int32_t curMoveYMax = std::numeric_limits<int32_t>::min();
            int32_t curMoveEmpty = 0;
            int32_t curMoveNeighbors = 0;
            vector<int32_t> curMoveVec;
            for (auto member : game.m_units[game.m_curUnitIndex].m_members)
            {
                Cell cell = transformCell(game.m_units[game.m_curUnitIndex].m_pivot,
                                          move.m_pos,
                                          move.m_rot,
                                          member);

                curMoveYTotal += cell.m_y;
                curMoveYMax = std::max(curMoveYMax, cell.m_y);

                for (size_t x = 0; x < game.m_grid.getWidth(); x++)
                {
                    if (game.m_grid(x, cell.m_y) == 0)
                    {
                        curMoveEmpty++;
                    }
                }

                for (uint32_t iNeighbor = 0; iNeighbor < 6; iNeighbor++)
                {
                    Cell neighbor = getNeighbor(iNeighbor, cell);
                    if (neighbor.m_x < 0 ||
                        neighbor.m_x >= (int32_t)game.m_grid.getWidth() ||
                        neighbor.m_y < 0 ||
                        neighbor.m_y >= (int32_t)game.m_grid.getHeight() ||
                        game.m_grid(neighbor.m_x, neighbor.m_y) != 0)
                    {
                        curMoveNeighbors++;
                    }
                }
            }

            if (scoreFunc == 0)
            {
                curMoveVec.push_back(curMoveYMax);
                curMoveVec.push_back(curMoveNeighbors);
            }
            else if (scoreFunc == 1)
            {
                curMoveVec.push_back(-curMoveEmpty + curMoveYMax);
                curMoveVec.push_back(curMoveNeighbors);
            }

            bool better = false;
            bool same = false;
            if (i == 0)
            {
                better = true;
            }
            else
            {
                if (std::lexicographical_compare(
                        bestMoveVec.begin(),
                        bestMoveVec.end(),
                        curMoveVec.begin(),
                        curMoveVec.end()))
                {
                    better = true;
                }
                else if (!std::lexicographical_compare(
                             curMoveVec.begin(),
                             curMoveVec.end(),
                             bestMoveVec.begin(),
                             bestMoveVec.end()))
                {
                    same = true;
                }
            }

            if (better)
            {
                bestMove = i;
                bestMoveVec = curMoveVec;
                sameCount = 1;
            }
            else if (same)
            {
                sameCount++;
                if (scoreSource.next() % sameCount == 0)
                {
                    bestMove = i;
                    bestMoveVec = curMoveVec;
                }
            }
        }

        if (!game.m_phrases.empty() && pPhraseSolver)
        {
            PhraseInterface pi(game);
            pi.m_endStates.push_back(moves[bestMove]);
            if (pPhraseSolver->wantTransformCache())
            {
                pi.m_transformCache.resize(6);
                for (uint32_t iCommand = 0; iCommand < 6; iCommand++)
                {
                    Command command = Command(iCommand + 1);
                    auto& curTable = pi.m_transformCache[iCommand];
                    curTable.init(
                        game.m_grid.getWidth(),
                        game.m_grid.getHeight(),
                        game.m_unitInfo[game.m_curUnitIndex].m_rots,
                        game.m_unitInfo[game.m_curUnitIndex].m_xMins,
                        game.m_unitInfo[game.m_curUnitIndex].m_yMins);
                    for (uint32_t rot = 0; rot < pi.getRots(); rot++)
                    {
                        for (int32_t y = 0; y < pi.getHeight(); y++)
                        {
                            int32_t yReal = y + pi.getYMin(rot);
                            for (int32_t x = 0; x < pi.getWidth(); x++)
                            {
                                int32_t xReal = x + pi.getXMin(rot);
                                UnitState curState(xReal, yReal, rot);
                                UnitState newState =
                                    game.transformUnit(command,
                                                       game.m_curUnitIndex,
                                                       curState);
                                curTable(curState) =
                                    game.isValid(game.m_curUnitIndex, newState) ?
                                    newState :
                                    UnitState(0, 0, -1);
                            }
                        }
                    }
                }
            }
            pPhraseSolver->processFrame(pi);
        }

        vector<Command> moveCommands;
        moveFinder.getCommands(moves[bestMove], &moveCommands);

        for (auto command : moveCommands)
        {
            char ch = commandToChar(command);
            game.step(ch);
        }
    }

    auto& solution = *pSolution;
    solution.m_problemId = problem.m_id;
    solution.m_seed = seed;
    solution.m_tag = "v9";

    if (!game.m_phrases.empty() && pPhraseSolver)
    {
        solution.m_chars.clear();
        pPhraseSolver->getChars(&solution.m_chars);

        calcScore(problem,
                  phrases,
                  seed,
                  solution,
                  pScore);
        if (*pScore < game.m_score)
        {
            solution.m_chars = game.m_chars;
            *pScore = game.m_score;
        }
    }
    else
    {
        solution.m_chars = game.m_chars;
        *pScore = game.m_score;
    }

    return true;
}
