#include "Common.hpp"
#include "Problem.hpp"
#include "Solution.hpp"
#include "Parser.hpp"
#include "FileUtils.hpp"
#include "CalcScore.hpp"
#include "Solver.hpp"
#include "Utils.hpp"

using std::string;
using std::vector;
using std::unique_ptr;

void usage()
{
    printf("Usage:\n");
    printf("  solve [options]\n");
    printf("Options:\n");
    printf("  -f <filename>     File containing JSON encoded input\n");
    printf("  -t <number>       Time limit, in seconds, to produce output\n");
    printf("  -m <number>       Memory limit, in megabytes, to produce output\n");
    printf("  -c <number>       Number of processor cores available\n");
    printf("  -p <string>       Phrase of power\n");
    printf("\n");
    printf("  -F <filename>...  Arguments until the next option are problem files\n");
    printf("  -P <filename>     File containing phrases of power on separate lines\n");
    printf("  -S <filename>     File containing best solutions found so far\n");
    printf("  -V                Verbose output to stderr\n");
}

int main(int argc, char* argv[])
{
    vector<string> argFiles;
    bool gotTime = false;
    uint64_t argTime = 0;
    bool gotMem = false;
    uint32_t argMem = 0;
    bool gotCores = false;
    uint32_t argCores = 0;
    vector<string> argPhrases;
    bool gotAllFiles = false;
    vector<string> argPhraseFiles;
    bool gotSolutionsFile = false;
    string argSolutionsFile;
    bool gotVerbose = false;

    (void)argTime;
    (void)argMem;
    (void)argCores;

    int iArg = 1;
    while (iArg < argc)
    {
        string strArg = argv[iArg];
        iArg++;

        if (strcmp(strArg.c_str(), "-f") == 0)
        {
            if (iArg >= argc)
            {
                usage();
                return 1;
            }
            argFiles.push_back(string(argv[iArg]));
            iArg++;
            gotAllFiles = false;
            continue;
        }

        if (strcmp(strArg.c_str(), "-t") == 0)
        {
            if (gotTime || iArg >= argc)
            {
                usage();
                return 1;
            }
            gotTime = true;
            argTime = atol(argv[iArg]);
            iArg++;
            gotAllFiles = false;
            continue;
        }

        if (strcmp(strArg.c_str(), "-m") == 0)
        {
            if (gotMem || iArg >= argc)
            {
                usage();
                return 1;
            }
            gotMem = true;
            argMem = atol(argv[iArg]);
            iArg++;
            gotAllFiles = false;
            continue;
        }

        if (strcmp(strArg.c_str(), "-c") == 0)
        {
            if (gotCores || iArg >= argc)
            {
                usage();
                return 1;
            }
            gotCores = true;
            argCores = atol(argv[iArg]);
            iArg++;
            gotAllFiles = false;
            continue;
        }

        if (strcmp(strArg.c_str(), "-p") == 0)
        {
            if (iArg >= argc)
            {
                usage();
                return 1;
            }
            argPhrases.push_back(string(argv[iArg]));
            iArg++;
            gotAllFiles = false;
            continue;
        }

        if (strcmp(strArg.c_str(), "-F") == 0)
        {
            gotAllFiles = true;
            continue;
        }

        if (strcmp(strArg.c_str(), "-P") == 0)
        {
            if (iArg >= argc)
            {
                usage();
                return 1;
            }
            argPhraseFiles.push_back(string(argv[iArg]));
            iArg++;
            gotAllFiles = false;
            continue;
        }

        if (strcmp(strArg.c_str(), "-S") == 0)
        {
            if (gotSolutionsFile || iArg >= argc)
            {
                usage();
                return 1;
            }
            gotSolutionsFile = true;
            argSolutionsFile = string(argv[iArg]);
            iArg++;
            gotAllFiles = false;
            continue;
        }

        if (strcmp(strArg.c_str(), "-V") == 0)
        {
            gotVerbose = true;
            continue;
        }

        if (!strArg.empty() && strArg[0] == '-')
        {
            usage();
            return 1;
        }

        if (gotAllFiles)
        {
            argFiles.push_back(strArg);
            continue;
        }

        usage();
        return 1;
    }

    if (argFiles.empty())
    {
        usage();
        return 1;
    }

    uint64_t startMS = getTicksMS();
    uint64_t stopMS = startMS + 1000000000;
    if (gotTime)
    {
        if (argTime <= 2)
        {
            stopMS = startMS + (argTime * 500);
        }
        else if (argTime <= 5)
        {
            stopMS = startMS + (argTime - 1) * 1000;
        }
        else if (argTime <= 30)
        {
            stopMS = startMS + (argTime - 2) * 1000;
        }
        else if (argTime <= 60)
        {
            stopMS = startMS + (argTime - 3) * 1000;
        }
        else if (argTime <= 300)
        {
            stopMS = startMS + (argTime - 5) * 1000;
        }
        else
        {
            stopMS = startMS + (argTime - 10) * 1000;
        }
    }

    string msg;

    vector<string> phrases(argPhrases);
    for (auto& strPhraseFile : argPhraseFiles)
    {
        FILE* f = fopen(strPhraseFile.c_str(), "rb");
        if (!f)
        {
            printf("Error opening file %s\n", strPhraseFile.c_str());
            return 1;
        }

        vector<string> lines;
        readLines(f, &lines);
        fclose(f);
        phrases.insert(phrases.end(), lines.begin(), lines.end());
    }

    vector<Problem> problems;
    for (auto& strProblemFile : argFiles)
    {
        problems.emplace_back();
        auto& problem = problems.back();

        if (!loadProblem(strProblemFile, &problem, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }
    }

    vector<Solution> prevSolutions;
    if (gotSolutionsFile)
    {
        if (!loadSolutions(argSolutionsFile, &prevSolutions, &msg))
        {
            printf("%s\n", msg.c_str());
            return 1;
        }
    }

    typedef std::pair<uint32_t, uint32_t> Inst;
    vector<Inst> insts;

    for (uint32_t iProblem = 0; iProblem < problems.size(); iProblem++)
    {
        auto& problem = problems[iProblem];
        for (uint32_t iSeed = 0; iSeed < problem.m_sourceSeeds.size(); iSeed++)
        {
            uint32_t seed = problem.m_sourceSeeds[iSeed];
            insts.push_back(Inst(iProblem, seed));
        }
    }

    std::sort(insts.begin(), insts.end(),
              [&](const Inst& inst1, const Inst& inst2) -> bool {
                  auto& problem1 = problems[inst1.first];
                  auto& problem2 = problems[inst2.first];
                  uint64_t cost1 =
                      (uint64_t)problem1.m_width *
                      (uint64_t)problem1.m_height *
                      (uint64_t)problem1.m_sourceLength;
                  uint64_t cost2 =
                      (uint64_t)problem2.m_width *
                      (uint64_t)problem2.m_height *
                      (uint64_t)problem2.m_sourceLength;
                  if (cost1 != cost2)
                  {
                      return cost1 < cost2;
                  }
                  if (problem1.m_id != problem2.m_id)
                  {
                      return problem1.m_id < problem2.m_id;
                  }
                  if (inst1.second != inst2.second)
                  {
                      return inst1.second < inst2.second;
                  }
                  return false;
              });

    vector<uint32_t> noImprovementCounts;
    noImprovementCounts.resize(insts.size());
    vector<uint32_t> bestScoreFuncs;
    bestScoreFuncs.resize(insts.size());
    vector<uint32_t> bestScoreSeeds;
    bestScoreSeeds.resize(insts.size());
    for (uint32_t iInst = 0; iInst < insts.size(); iInst++)
    {
        noImprovementCounts[iInst] = 0;
        bestScoreFuncs[iInst] = 0;
        bestScoreSeeds[iInst] = 0;
    }

    vector<Solution> solutions;
    solutions.resize(insts.size());

    vector<uint64_t> scores;
    scores.resize(insts.size());

    for (uint32_t iInst = 0; iInst < insts.size(); iInst++)
    {
        auto& inst = insts[iInst];
        uint32_t iProblem = inst.first;
        auto& problem = problems[iProblem];
        uint32_t seed = inst.second;

        size_t iPrevSolution = 0;
        for (iPrevSolution = 0; iPrevSolution < prevSolutions.size(); iPrevSolution++)
        {
            if (prevSolutions[iPrevSolution].m_problemId == problem.m_id &&
                prevSolutions[iPrevSolution].m_seed == seed)
            {
                break;
            }
        }
        if (iPrevSolution == prevSolutions.size())
        {
            continue;
        }
        const Solution& solution = prevSolutions[iPrevSolution];

        uint64_t score = 0;
        calcScore(problem, phrases, seed, solution, &score);

        solutions[iInst] = solution;
        scores[iInst] = score;
    }

    bool timeOut = false;

    auto solveOne = [&](uint32_t iInst,
                        uint32_t iPhraseSolver,
                        uint32_t scoreFunc,
                        uint32_t scoreSeed)
    {
        auto& inst = insts[iInst];
        uint32_t iProblem = inst.first;
        auto& problem = problems[iProblem];
        uint32_t seed = inst.second;

        Solution solution;
        uint64_t score = 0;
        if (solve(problem, phrases,
                  iPhraseSolver,
                  seed, scoreFunc, scoreSeed, stopMS,
                  &solution, &score))
        {
            if (gotVerbose)
            {
                fprintf(stderr,
                        "(%2" PRId64 ", %5" PRIu32 "): %6" PRIu64 " -> %6" PRIu64 "%s\n",
                        problem.m_id,
                        seed,
                        scores[iInst],
                        score,
                        score > scores[iInst] ? "   BETTER!" :
                        score < scores[iInst] ? "   (worse)" : "");
            }
            if (score > scores[iInst])
            {
                solutions[iInst] = std::move(solution);
                scores[iInst] = score;

                noImprovementCounts[iInst] = 0;
                bestScoreFuncs[iInst] = scoreFunc;
                bestScoreSeeds[iInst] = scoreSeed;
            }
            else
            {
                noImprovementCounts[iInst]++;
            }
        }

        if (stopMS != 0 && getTicksMS() >= stopMS)
        {
            timeOut = true;
        }
    };

    for (uint32_t scoreSeed = 0; !timeOut && scoreSeed < 1000; scoreSeed++)
    for (uint32_t iPhraseSolver = 2; !timeOut && iPhraseSolver < 3; iPhraseSolver++)
    for (uint32_t scoreFunc = 0; !timeOut && scoreFunc < 2; scoreFunc++)
    {
        bool didSomething = false;

        for (uint32_t iInst = 0; !timeOut && iInst < insts.size(); iInst++)
        {
            if (noImprovementCounts[iInst] >= 30)
            {
                continue;
            }
            didSomething = true;

            solveOne(iInst, iPhraseSolver, scoreFunc, scoreSeed);
        }

        if (!didSomething)
        {
            goto nextAlgorithm; // Yes, goto!
        }
    }

nextAlgorithm:

    for (uint32_t iInst = 0; !timeOut && iInst < insts.size(); iInst++)
    {
        const uint32_t iPhraseSolver = 3;
        solveOne(iInst,
                 iPhraseSolver,
                 bestScoreFuncs[iInst],
                 bestScoreSeeds[iInst]);
    }

    for (uint32_t scoreSeed = 0; !timeOut && scoreSeed < 1000; scoreSeed++)
    for (uint32_t scoreFunc = 0; !timeOut && scoreFunc < 2; scoreFunc++)
    for (uint32_t iInst = 0; !timeOut && iInst < insts.size(); iInst++)
    {
        if (scoreFunc == bestScoreFuncs[iInst] &&
            scoreSeed == bestScoreSeeds[iInst])
        {
            continue;
        }

        const uint32_t iPhraseSolver = 3;
        solveOne(iInst,
                 iPhraseSolver,
                 scoreFunc,
                 scoreSeed);
    }

    vector<Solution> positiveSolutions;
    for (uint32_t iInst = 0; iInst < insts.size(); iInst++)
    {
        if (scores[iInst] > 0)
        {
            positiveSolutions.push_back(std::move(solutions[iInst]));
        }
    }

    string strSolutions;
    if (!formatSolutions(positiveSolutions, &strSolutions, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    printf("%s", strSolutions.c_str());

    return 0;
}
